@@ -2,195 +2,23 @@
 //! which wrap native Python objects to provide additional functionality
 //! or tighter integration with de/compression algorithms.
 //!
+use std::convert::TryFrom;
 use std::fs::{File, OpenOptions};
 use std::io::{copy, Cursor, Read, Seek, SeekFrom, Write};
+use std::mem;
+use std::os::raw::c_int;
 
 use crate::exceptions::CompressionError;
 use crate::BytesType;
-use numpy::PyArray1;
-use pyo3::class::buffer::PyBufferProtocol;
+use pyo3::exceptions::{self, PyBufferError};
+use pyo3::ffi;
 use pyo3::prelude::*;
-use pyo3::types::{PyByteArray, PyBytes};
-use pyo3::{ffi, PySequenceProtocol};
-use pyo3::{AsPyPointer, PyObjectProtocol};
+use pyo3::types::PyBytes;
 use std::path::PathBuf;
 
 pub(crate) trait AsBytes {
     fn as_bytes(&self) -> &[u8];
-    fn as_bytes_mut(&mut self) -> &mut [u8];
-}
-
-/// Internal wrapper for `numpy.array`/`PyArray1`, to provide Read + Write and other traits
-pub struct RustyNumpyArray<'a> {
-    pub(crate) inner: &'a PyArray1<u8>,
-    pub(crate) cursor: Cursor<&'a mut [u8]>,
-}
-impl<'a> AsBytes for RustyNumpyArray<'a> {
-    fn as_bytes(&self) -> &[u8] {
-        self.cursor.get_ref().as_ref()
-    }
-    fn as_bytes_mut(&mut self) -> &mut [u8] {
-        self.cursor.get_mut()
-    }
-}
-impl<'a> RustyNumpyArray<'a> {
-    pub(crate) fn as_bytes(&self) -> &[u8] {
-        unsafe { self.inner.as_slice().unwrap() }
-    }
-}
-impl<'a> From<&'a PyArray1<u8>> for RustyNumpyArray<'a> {
-    fn from(inner: &'a PyArray1<u8>) -> Self {
-        Self {
-            inner,
-            cursor: Cursor::new(unsafe { inner.as_slice_mut().unwrap() }),
-        }
-    }
-}
-impl<'a> FromPyObject<'a> for RustyNumpyArray<'a> {
-    fn extract(ob: &'a PyAny) -> PyResult<Self> {
-        let pybytes: &PyArray1<u8> = ob.extract()?;
-        Ok(Self::from(pybytes))
-    }
-}
-impl<'a> ToPyObject for RustyNumpyArray<'a> {
-    fn to_object(&self, py: Python<'_>) -> PyObject {
-        self.inner.to_object(py)
-    }
-}
-impl<'a> Write for RustyNumpyArray<'a> {
-    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
-        self.cursor.write(buf)
-    }
-    fn flush(&mut self) -> std::io::Result<()> {
-        self.cursor.flush()
-    }
-}
-impl<'a> Read for RustyNumpyArray<'a> {
-    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
-        self.cursor.read(buf)
-    }
-}
-
-impl<'a> Seek for RustyNumpyArray<'a> {
-    fn seek(&mut self, pos: SeekFrom) -> std::io::Result<u64> {
-        self.cursor.seek(pos)
-    }
-}
-
-/// Internal wrapper for `bytes`/`PyBytes`, to provide Read + Write and other traits
-pub struct RustyPyBytes<'a> {
-    pub(crate) inner: &'a PyBytes,
-    pub(crate) cursor: Cursor<&'a mut [u8]>,
-}
-impl<'a> AsBytes for RustyPyBytes<'a> {
-    fn as_bytes(&self) -> &[u8] {
-        self.inner.as_bytes()
-    }
-    fn as_bytes_mut(&mut self) -> &mut [u8] {
-        self.cursor.get_mut()
-    }
-}
-impl<'a> From<&'a PyBytes> for RustyPyBytes<'a> {
-    fn from(inner: &'a PyBytes) -> Self {
-        let ptr = inner.as_bytes().as_ptr();
-        Self {
-            inner,
-            cursor: Cursor::new(unsafe { std::slice::from_raw_parts_mut(ptr as *mut _, inner.as_bytes().len()) }),
-        }
-    }
-}
-impl<'a> FromPyObject<'a> for RustyPyBytes<'a> {
-    fn extract(ob: &'a PyAny) -> PyResult<Self> {
-        let pybytes: &PyBytes = ob.extract()?;
-        Ok(Self::from(pybytes))
-    }
-}
-impl<'a> ToPyObject for RustyPyBytes<'a> {
-    fn to_object(&self, py: Python<'_>) -> PyObject {
-        self.inner.to_object(py)
-    }
-}
-impl<'a> Read for RustyPyBytes<'a> {
-    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
-        self.cursor.read(buf)
-    }
-}
-impl<'a> Write for RustyPyBytes<'a> {
-    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
-        self.cursor.write(buf)
-    }
-    fn flush(&mut self) -> std::io::Result<()> {
-        self.cursor.flush()
-    }
-}
-impl<'a> Seek for RustyPyBytes<'a> {
-    fn seek(&mut self, style: SeekFrom) -> std::io::Result<u64> {
-        self.cursor.seek(style)
-    }
-}
-
-/// Internal wrapper for `bytearray`/`PyByteArray`, to provide Read + Write and other traits
-pub struct RustyPyByteArray<'a> {
-    pub(crate) inner: &'a PyByteArray,
-    pub(crate) cursor: Cursor<&'a mut [u8]>,
-}
-impl<'a> AsBytes for RustyPyByteArray<'a> {
-    fn as_bytes(&self) -> &[u8] {
-        self.cursor.get_ref()
-    }
-    fn as_bytes_mut(&mut self) -> &mut [u8] {
-        self.cursor.get_mut()
-    }
-}
-impl<'a> From<&'a PyByteArray> for RustyPyByteArray<'a> {
-    fn from(inner: &'a PyByteArray) -> Self {
-        Self {
-            inner,
-            cursor: Cursor::new(unsafe { inner.as_bytes_mut() }),
-        }
-    }
-}
-impl<'a> FromPyObject<'a> for RustyPyByteArray<'a> {
-    fn extract(ob: &'a PyAny) -> PyResult<Self> {
-        let pybytes: &PyByteArray = ob.extract()?;
-        Ok(Self::from(pybytes))
-    }
-}
-impl<'a> ToPyObject for RustyPyByteArray<'a> {
-    fn to_object(&self, py: Python<'_>) -> PyObject {
-        self.inner.to_object(py)
-    }
-}
-impl<'a> Write for RustyPyByteArray<'a> {
-    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
-        if (self.cursor.position() as usize + buf.len()) > self.inner.len() {
-            let previous_pos = self.cursor.position();
-            self.inner.resize(self.cursor.position() as usize + buf.len()).unwrap();
-            self.cursor = Cursor::new(unsafe { self.inner.as_bytes_mut() });
-            self.cursor.set_position(previous_pos);
-        }
-        self.cursor.write(buf)
-    }
-    fn flush(&mut self) -> std::io::Result<()> {
-        if self.inner.len() != self.cursor.position() as usize {
-            let prev_pos = self.cursor.position();
-            self.inner.resize(self.cursor.position() as usize).unwrap();
-            self.cursor = Cursor::new(unsafe { self.inner.as_bytes_mut() });
-            self.cursor.set_position(prev_pos);
-        }
-        Ok(())
-    }
-}
-impl<'a> Read for RustyPyByteArray<'a> {
-    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
-        self.cursor.read(buf)
-    }
-}
-
-impl<'a> Seek for RustyPyByteArray<'a> {
-    fn seek(&mut self, pos: SeekFrom) -> std::io::Result<u64> {
-        self.cursor.seek(pos)
-    }
+    fn as_bytes_mut(&mut self) -> PyResult<&mut [u8]>;
 }
 
 /// A native Rust file-like object. Reading and writing takes place
@@ -214,28 +42,6 @@ pub struct RustyFile {
     pub(crate) inner: File,
 }
 
-#[pyproto]
-impl PyObjectProtocol for RustyFile {
-    fn __repr__(&self) -> PyResult<String> {
-        let path = match self.path.as_path().to_str() {
-            Some(path) => path.to_string(),
-            None => self.path.to_string_lossy().to_string(),
-        };
-        let repr = format!("cramjam.File(path={}, len={:?})", path, self.len()?);
-        Ok(repr)
-    }
-    fn __bool__(&self) -> PyResult<bool> {
-        Ok(self.len()? > 0)
-    }
-}
-
-#[pyproto]
-impl PySequenceProtocol for RustyFile {
-    fn __len__(&self) -> PyResult<usize> {
-        self.len()
-    }
-}
-
 impl AsBytes for RustyFile {
     fn as_bytes(&self) -> &[u8] {
         unimplemented!(
@@ -243,7 +49,7 @@ impl AsBytes for RustyFile {
         entire file into memory; consider using cramjam.Buffer"
         )
     }
-    fn as_bytes_mut(&mut self) -> &mut [u8] {
+    fn as_bytes_mut(&mut self) -> PyResult<&mut [u8]> {
         unimplemented!(
             "Converting a File to bytes is not supported, as it'd require reading the \
         entire file into memory; consider using cramjam.Buffer"
@@ -263,6 +69,7 @@ impl RustyFile {
     /// b'tes'
     /// ```
     #[new]
+    #[pyo3(signature = (path, read = None, write = None, truncate = None, append = None))]
     pub fn __init__(
         path: &str,
         read: Option<bool>,
@@ -288,7 +95,8 @@ impl RustyFile {
     }
     /// Read from the file in its current position, returns `bytes`; optionally specify number of
     /// bytes to read.
-    pub fn read<'a>(&mut self, py: Python<'a>, n_bytes: Option<usize>) -> PyResult<&'a PyBytes> {
+    #[pyo3(signature = (n_bytes=None))]
+    pub fn read<'a>(&mut self, py: Python<'a>, n_bytes: Option<usize>) -> PyResult<Bound<'a, PyBytes>> {
         read(self, py, n_bytes)
     }
     /// Read from the file in its current position, into a [`BytesType`](../enum.BytesType.html) object.
@@ -303,6 +111,7 @@ impl RustyFile {
     /// 1: from current stream position
     /// 2: from end of the stream
     /// ```
+    #[pyo3(signature = (position, whence=None))]
     pub fn seek(&mut self, position: isize, whence: Option<usize>) -> PyResult<usize> {
         let pos = match whence.unwrap_or_else(|| 0) {
             0 => SeekFrom::Start(position as u64),
@@ -344,6 +153,175 @@ impl RustyFile {
             .map_err(|e| pyo3::exceptions::PyOSError::new_err(e.to_string()))?;
         Ok(meta.len() as usize)
     }
+
+    fn __repr__(&self) -> PyResult<String> {
+        let path = match self.path.as_path().to_str() {
+            Some(path) => path.to_string(),
+            None => self.path.to_string_lossy().to_string(),
+        };
+        let repr = format!("cramjam.File<path={}, len={:?}>", path, self.len()?);
+        Ok(repr)
+    }
+    fn __bool__(&self) -> PyResult<bool> {
+        Ok(self.len()? > 0)
+    }
+    fn __len__(&self) -> PyResult<usize> {
+        self.len()
+    }
+}
+
+/// Internal wrapper to PyBuffer, not exposed thru API
+/// used only for impl of Read/Write
+// Inspired from pyo3 PyBuffer<T>, but here we don't want or care about T
+pub struct PythonBuffer {
+    pub(crate) inner: std::pin::Pin<Box<ffi::Py_buffer>>,
+    pub(crate) pos: usize,
+    #[cfg(PyPy)]
+    pub(crate) owner: PyObject,
+}
+// PyBuffer is thread-safe: the shape of the buffer is immutable while a Py_buffer exists.
+// Accessing the buffer contents is protected using the GIL.
+unsafe impl Send for PythonBuffer {}
+unsafe impl Sync for PythonBuffer {}
+
+impl PythonBuffer {
+    /// Reset the read/write position of cursor
+    pub fn reset_position(&mut self) {
+        self.pos = 0;
+    }
+    /// Explicitly set the position of the cursor
+    pub fn set_position(&mut self, pos: usize) {
+        self.pos = pos;
+    }
+    /// Get the current position of the cursor
+    pub fn position(&self) -> usize {
+        self.pos
+    }
+    /// Is the Python buffer readonly
+    pub fn readonly(&self) -> bool {
+        self.inner.readonly == 1
+    }
+    /// Get the underlying buffer as a slice of bytes
+    pub fn as_slice(&self) -> &[u8] {
+        unsafe { std::slice::from_raw_parts(self.buf_ptr() as *const u8, self.len_bytes()) }
+    }
+    /// Get the underlying buffer as a mutable slice of bytes
+    pub fn as_slice_mut(&mut self) -> PyResult<&mut [u8]> {
+        #[cfg(PyPy)]
+        {
+            Python::with_gil(|py| {
+                let is_memoryview = unsafe { ffi::PyMemoryView_Check(self.owner.as_ptr()) } == 1;
+                if is_memoryview || self.owner.bind(py).is_instance_of::<PyBytes>() {
+                    Err(pyo3::exceptions::PyTypeError::new_err(
+                        "With PyPy, an output of type `bytes` or `memoryview` does not work. See issue pypy/pypy#4918",
+                    ))
+                } else {
+                    Ok(())
+                }
+            })?;
+        }
+        Ok(unsafe { std::slice::from_raw_parts_mut(self.buf_ptr() as *mut u8, self.len_bytes()) })
+    }
+    /// If underlying buffer is c_contiguous
+    pub fn is_c_contiguous(&self) -> bool {
+        unsafe { ffi::PyBuffer_IsContiguous(&*self.inner as *const ffi::Py_buffer, b'C' as std::os::raw::c_char) == 1 }
+    }
+    /// Dimensions for buffer
+    pub fn dimensions(&self) -> usize {
+        self.inner.ndim as usize
+    }
+    /// raw pointer to buffer
+    pub fn buf_ptr(&self) -> *mut std::os::raw::c_void {
+        self.inner.buf
+    }
+    /// length of buffer in bytes
+    pub fn len_bytes(&self) -> usize {
+        self.inner.len as usize
+    }
+    /// the buffer item size
+    pub fn item_size(&self) -> usize {
+        self.inner.itemsize as usize
+    }
+    /// number of items in buffer
+    pub fn item_count(&self) -> usize {
+        (self.inner.len as usize) / (self.inner.itemsize as usize)
+    }
+}
+
+impl Drop for PythonBuffer {
+    fn drop(&mut self) {
+        Python::with_gil(|_| unsafe { ffi::PyBuffer_Release(&mut *self.inner) })
+    }
+}
+
+impl<'py> FromPyObject<'py> for PythonBuffer {
+    fn extract_bound(obj: &Bound<'py, PyAny>) -> PyResult<Self> {
+        Self::try_from(obj)
+    }
+}
+
+impl<'a, 'py> TryFrom<&'a Bound<'py, PyAny>> for PythonBuffer {
+    type Error = PyErr;
+    fn try_from(obj: &'a Bound<'py, PyAny>) -> Result<Self, Self::Error> {
+        let mut buf = Box::new(mem::MaybeUninit::uninit());
+        let rc = unsafe { ffi::PyObject_GetBuffer(obj.as_ptr(), buf.as_mut_ptr(), ffi::PyBUF_CONTIG_RO) };
+        if rc != 0 {
+            return Err(exceptions::PyBufferError::new_err(
+                "Failed to get buffer, is it C contiguous, and shape is not null?",
+            ));
+        }
+        let buf = Box::new(unsafe { mem::MaybeUninit::<ffi::Py_buffer>::assume_init(*buf) });
+        let buf = Self {
+            inner: std::pin::Pin::from(buf),
+            pos: 0,
+            #[cfg(PyPy)]
+            owner: Python::with_gil(|py| obj.to_object(py)),
+        };
+        // sanity checks
+        if buf.inner.shape.is_null() {
+            Err(exceptions::PyBufferError::new_err("shape is null"))
+        } else if !buf.is_c_contiguous() {
+            Err(PyBufferError::new_err("Buffer is not C contiguous"))
+        } else {
+            Ok(buf)
+        }
+    }
+}
+
+impl Read for PythonBuffer {
+    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
+        let slice = self.as_slice();
+        if self.pos < slice.len() {
+            let nbytes = (&slice[self.pos..]).read(buf)?;
+            self.pos += nbytes;
+            Ok(nbytes)
+        } else {
+            Ok(0)
+        }
+    }
+}
+
+impl Write for PythonBuffer {
+    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
+        let pos = self.position();
+        let slice = self
+            .as_slice_mut()
+            .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, e.to_string()))?;
+        let len = slice.len();
+
+        if pos < slice.len() {
+            let nbytes = std::cmp::min(len - pos, buf.len());
+            slice[pos..pos + nbytes].copy_from_slice(&buf[..nbytes]);
+            self.pos += nbytes;
+            Ok(nbytes)
+        } else {
+            Ok(0)
+        }
+    }
+
+    fn flush(&mut self) -> std::io::Result<()> {
+        Ok(())
+    }
 }
 
 /// A native Rust file-like object. Reading and writing takes place
@@ -358,7 +336,7 @@ impl RustyFile {
 /// b'bytes'
 /// ```
 ///
-#[pyclass(name = "Buffer")]
+#[pyclass(subclass, name = "Buffer")]
 #[derive(Default)]
 pub struct RustyBuffer {
     pub(crate) inner: Cursor<Vec<u8>>,
@@ -368,8 +346,9 @@ impl AsBytes for RustyBuffer {
     fn as_bytes(&self) -> &[u8] {
         self.inner.get_ref().as_slice()
     }
-    fn as_bytes_mut(&mut self) -> &mut [u8] {
-        self.inner.get_mut().as_mut_slice()
+    fn as_bytes_mut(&mut self) -> PyResult<&mut [u8]> {
+        let slice = self.inner.get_mut().as_mut_slice();
+        Ok(slice)
     }
 }
 
@@ -379,75 +358,6 @@ impl From<Vec<u8>> for RustyBuffer {
     }
 }
 
-#[pyproto]
-impl PyBufferProtocol for RustyBuffer {
-    fn bf_getbuffer(slf: PyRefMut<Self>, view: *mut ffi::Py_buffer, flags: std::os::raw::c_int) -> PyResult<()> {
-        if view.is_null() {
-            return Err(pyo3::exceptions::PyBufferError::new_err("View is null"));
-        }
-
-        if (flags & ffi::PyBUF_WRITABLE) == ffi::PyBUF_WRITABLE {
-            return Err(pyo3::exceptions::PyBufferError::new_err("Object is not writable"));
-        }
-
-        unsafe {
-            (*view).obj = slf.as_ptr();
-            ffi::Py_INCREF((*view).obj);
-        }
-
-        let bytes = slf.inner.get_ref().as_slice();
-
-        unsafe {
-            (*view).buf = bytes.as_ptr() as *mut std::os::raw::c_void;
-            (*view).len = bytes.len() as isize;
-            (*view).readonly = 1;
-            (*view).itemsize = 1;
-
-            (*view).format = std::ptr::null_mut();
-            if (flags & ffi::PyBUF_FORMAT) == ffi::PyBUF_FORMAT {
-                let msg = std::ffi::CStr::from_bytes_with_nul(b"B\0").unwrap();
-                (*view).format = msg.as_ptr() as *mut _;
-            }
-
-            (*view).ndim = 1;
-            (*view).shape = std::ptr::null_mut();
-            if (flags & ffi::PyBUF_ND) == ffi::PyBUF_ND {
-                (*view).shape = (&((*view).len)) as *const _ as *mut _;
-            }
-
-            (*view).strides = std::ptr::null_mut();
-            if (flags & ffi::PyBUF_STRIDES) == ffi::PyBUF_STRIDES {
-                (*view).strides = &((*view).itemsize) as *const _ as *mut _;
-            }
-
-            (*view).suboffsets = std::ptr::null_mut();
-            (*view).internal = std::ptr::null_mut();
-        }
-        Ok(())
-    }
-    fn bf_releasebuffer(_slf: PyRefMut<Self>, _view: *mut ffi::Py_buffer) {}
-}
-
-#[pyproto]
-impl PySequenceProtocol for RustyBuffer {
-    fn __len__(&self) -> usize {
-        self.len()
-    }
-    fn __contains__(&self, x: u8) -> bool {
-        self.inner.get_ref().contains(&x)
-    }
-}
-
-#[pyproto]
-impl PyObjectProtocol for RustyBuffer {
-    fn __repr__(&self) -> String {
-        format!("cramjam.Buffer(len={:?})", self.len())
-    }
-    fn __bool__(&self) -> bool {
-        self.len() > 0
-    }
-}
-
 /// A Buffer object, similar to [cramjam.File](struct.RustyFile.html) only the bytes are held in-memory
 ///
 /// ### Example
@@ -461,6 +371,7 @@ impl PyObjectProtocol for RustyBuffer {
 impl RustyBuffer {
     /// Instantiate the object, optionally with any supported bytes-like object in [BytesType](../enum.BytesType.html)
     #[new]
+    #[pyo3(signature = (data=None))]
     pub fn __init__(mut data: Option<BytesType<'_>>) -> PyResult<Self> {
         let mut buf = vec![];
         if let Some(bytes) = data.as_mut() {
@@ -482,7 +393,8 @@ impl RustyBuffer {
         Ok(r as usize)
     }
     /// Read from the buffer in its current position, returns bytes; optionally specify number of bytes to read.
-    pub fn read<'a>(&mut self, py: Python<'a>, n_bytes: Option<usize>) -> PyResult<&'a PyBytes> {
+    #[pyo3(signature = (n_bytes=None))]
+    pub fn read<'a>(&mut self, py: Python<'a>, n_bytes: Option<usize>) -> PyResult<Bound<'a, PyBytes>> {
         read(self, py, n_bytes)
     }
     /// Read from the buffer in its current position, into a [BytesType](../enum.BytesType.html) object.
@@ -496,6 +408,7 @@ impl RustyBuffer {
     /// 1: from current stream position
     /// 2: from end of the stream
     /// ```
+    #[pyo3(signature = (position, whence=None))]
     pub fn seek(&mut self, position: isize, whence: Option<usize>) -> PyResult<usize> {
         let pos = match whence.unwrap_or_else(|| 0) {
             0 => SeekFrom::Start(position as u64),
@@ -530,32 +443,85 @@ impl RustyBuffer {
         self.inner.set_position(0);
         Ok(())
     }
+
+    fn __len__(&self) -> usize {
+        self.len()
+    }
+    fn __contains__(&self, py: Python, x: BytesType) -> bool {
+        let bytes = x.as_bytes();
+        py.allow_threads(|| self.inner.get_ref().windows(bytes.len()).any(|w| w == bytes))
+    }
+    fn __repr__(&self) -> String {
+        format!("cramjam.Buffer<len={:?}>", self.len())
+    }
+    fn __eq__(&self, other: &Self) -> bool {
+        self.inner == other.inner
+    }
+    fn __bool__(&self) -> bool {
+        self.len() > 0
+    }
+    unsafe fn __getbuffer__(slf: PyRefMut<Self>, view: *mut ffi::Py_buffer, flags: c_int) -> PyResult<()> {
+        if view.is_null() {
+            return Err(pyo3::exceptions::PyBufferError::new_err("View is null"));
+        }
+
+        if (flags & ffi::PyBUF_WRITABLE) == ffi::PyBUF_WRITABLE {
+            return Err(pyo3::exceptions::PyBufferError::new_err("Object is not writable"));
+        }
+
+        (*view).obj = slf.as_ptr();
+        ffi::Py_INCREF((*view).obj);
+
+        let bytes = slf.inner.get_ref().as_slice();
+
+        (*view).buf = bytes.as_ptr() as *mut std::os::raw::c_void;
+        (*view).len = bytes.len() as isize;
+        (*view).readonly = 0;
+        (*view).itemsize = 1;
+
+        (*view).format = std::ptr::null_mut();
+        if (flags & ffi::PyBUF_FORMAT) == ffi::PyBUF_FORMAT {
+            let msg = std::ffi::CStr::from_bytes_with_nul(b"B\0").unwrap();
+            (*view).format = msg.as_ptr() as *mut _;
+        }
+
+        (*view).ndim = 1;
+        (*view).shape = std::ptr::null_mut();
+        if (flags & ffi::PyBUF_ND) == ffi::PyBUF_ND {
+            (*view).shape = (&((*view).len)) as *const _ as *mut _;
+        }
+
+        (*view).strides = std::ptr::null_mut();
+        if (flags & ffi::PyBUF_STRIDES) == ffi::PyBUF_STRIDES {
+            (*view).strides = &((*view).itemsize) as *const _ as *mut _;
+        }
+
+        (*view).suboffsets = std::ptr::null_mut();
+        (*view).internal = std::ptr::null_mut();
+        Ok(())
+    }
+    unsafe fn __releasebuffer__(&self, _view: *mut ffi::Py_buffer) {}
 }
 
 fn write<W: Write>(input: &mut BytesType, output: &mut W) -> std::io::Result<u64> {
     let result = match input {
+        BytesType::RustyBuffer(buf) => copy(&mut buf.borrow_mut().inner, output)?,
         BytesType::RustyFile(data) => copy(&mut data.borrow_mut().inner, output)?,
-        BytesType::RustyBuffer(data) => copy(&mut data.borrow_mut().inner, output)?,
-        BytesType::ByteArray(data) => copy(data, output)?,
-        BytesType::NumpyArray(array) => copy(array, output)?,
-        BytesType::Bytes(data) => {
-            let buffer = data.as_bytes();
-            copy(&mut Cursor::new(buffer), output)?
-        }
+        BytesType::PyBuffer(buf) => copy(buf, output)?,
     };
     Ok(result)
 }
 
-fn read<'a, R: Read>(reader: &mut R, py: Python<'a>, n_bytes: Option<usize>) -> PyResult<&'a PyBytes> {
+fn read<'a, R: Read>(reader: &mut R, py: Python<'a>, n_bytes: Option<usize>) -> PyResult<Bound<'a, PyBytes>> {
     match n_bytes {
-        Some(n) => PyBytes::new_with(py, n, |buf| {
+        Some(n) => PyBytes::new_bound_with(py, n, |buf| {
             reader.read(buf)?;
             Ok(())
         }),
         None => {
             let mut buf = vec![];
             reader.read_to_end(&mut buf)?;
-            Ok(PyBytes::new(py, buf.as_slice()))
+            Ok(PyBytes::new_bound(py, buf.as_slice()))
         }
     }
 }
@@ -570,6 +536,18 @@ impl Seek for RustyFile {
         self.inner.seek(pos)
     }
 }
+impl Seek for PythonBuffer {
+    fn seek(&mut self, pos: SeekFrom) -> std::io::Result<u64> {
+        let len = self.len_bytes();
+        let current = self.position();
+        match pos {
+            SeekFrom::Start(n) => self.set_position(n as usize),
+            SeekFrom::End(n) => self.set_position((len as i64 - n) as usize),
+            SeekFrom::Current(n) => self.set_position((current as i64 + n) as usize),
+        }
+        Ok(self.position() as _)
+    }
+}
 impl Write for RustyBuffer {
     fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
         self.inner.write(buf)
@@ -602,10 +580,9 @@ impl Read for RustyFile {
 #[inline(always)]
 pub(crate) fn stream_compress<W: Write>(encoder: &mut Option<W>, input: &[u8]) -> PyResult<usize> {
     match encoder {
-        Some(encoder) => {
-            let result = std::io::copy(&mut Cursor::new(input), encoder).map(|v| v as usize);
-            crate::to_py_err!(CompressionError -> result)
-        }
+        Some(encoder) => std::io::copy(&mut Cursor::new(input), encoder)
+            .map(|v| v as usize)
+            .map_err(CompressionError::from_err),
         None => Err(CompressionError::new_err(
             "Compressor looks to have been consumed via `finish()`. \
             please create a new compressor instance.",
@@ -629,7 +606,7 @@ where
 
     match detached_encoder {
         Some(encoder) => {
-            let result = crate::to_py_err!(CompressionError -> into_vec(encoder))?;
+            let result = into_vec(encoder).map_err(CompressionError::from_err)?;
             Ok(RustyBuffer::from(result))
         }
         None => Ok(RustyBuffer::from(vec![])),
@@ -645,7 +622,7 @@ where
 {
     match encoder {
         Some(inner) => {
-            crate::to_py_err!(CompressionError -> inner.flush())?;
+            inner.flush().map_err(CompressionError::from_err)?;
             let cursor = cursor_mut_ref(inner);
             let buf = RustyBuffer::from(cursor.get_ref().clone());
             cursor.get_mut().truncate(0);
